__version__ = "0.0.1"

from mccube.components import (
    AbstractPropagator,
    AbstractRecombinator,
    LangevinDiffusionPropagator,
    MonteCarloRecombinator,
    WrappedPropagator,
    WrappedRecombinator,
)
from mccube.cubature import MCCubatureStep, mccubaturesolve
from mccube.formulae import minimal_cubature_formula

__all__ = [
    # Base
    "mccubaturesolve",
    "MCCubatureStep",
    # Formulae
    "minimal_cubature_formula",
    # Components
    "AbstractPropagator",
    "AbstractRecombinator",
    "WrappedPropagator",
    "WrappedRecombinator",
    "LangevinDiffusionPropagator",
    "MonteCarloRecombinator",
]



"""Helpful utilities used throughout MCCube."""
from typing import Callable, Optional

import jax
import jax.tree_util as jtu
from jaxtyping import PyTree


def no_operation(*args, **kwds):
    ...


def identity_transform(t: float, particles: PyTree, args: PyTree) -> PyTree:
    return particles


def identity_validator(transform: Callable) -> Callable:
    return transform


# Below code is taken from Diffrax https://github.com/patrick-kidger/diffrax/blob/f101e75976e9dea86eb57f028f1f3bed646af1db/diffrax/misc.py#L155
def split_by_tree(key, tree, is_leaf: Optional[Callable[[PyTree], bool]] = None):
    """From Diffrax (https://github.com/patrick-kidger/diffrax/). Like jax.random.split
    but accepts tree as a second argument and produces a tree of keys with the same
    structure.
    """
    treedef = jtu.tree_structure(tree, is_leaf=is_leaf)
    return jtu.tree_unflatten(treedef, jax.random.split(key, treedef.num_leaves))



"""Base classes for creating and defining components."""

from __future__ import annotations

import abc
from typing import Callable, TypeVar
from typing_extensions import ParamSpec

import equinox as eqx
import jax.tree_util as jtu
from jaxtyping import Array, ArrayLike, Float, PyTree

from mccube.utils import identity_transform, identity_validator

P = ParamSpec("P")
T = TypeVar("T")


class AbstractComponent(eqx.Module):
    r"""Abstract base class for all Components.

    Given some particles $p(t)$, a Component defines the composition of a
    transformation $f(t, p(t), args)$, with a validation rule $v(f)$, resulting in
    the validated transform $h(t, p(t), args) = (v \circ f)(t, p(t), args)$.
    """

    @abc.abstractmethod
    def transform(
        self, time: float, particles: PyTree[Float[ArrayLike, "..."]], args: PyTree
    ) -> PyTree[Float[Array, "..."]]:
        r"""Transform the particles.

        Represents a function $f(t, p(t), args)$.

        Args:
            time: particle existance time.
            particles: particles to transform.
            args: additional static arguments passed to the transform.

        Returns:
            Transformed particles.
        """
        ...

    @abc.abstractmethod
    def validate(self, transform: Callable[P, T]) -> Callable[P, T]:
        r"""Validate the transform.

        Ensures that the transform obeys certain properties.

        Args:
            transform: transform to validate, $f(t, p(t), args)$.

        Returns:
            Validated transform $(v \circ f)(t, p(t), args)$.
        """
        ...

    def __call__(
        self, time: float, particles: PyTree[Float[ArrayLike, "..."]], args: PyTree
    ) -> PyTree[Float[Array, "..."]]:
        r"""Evaluate the validated transform.

        Args:
            time: particle existance time.
            particles: particles to transform with PyTree strucutre $P$.
            args: additional static arguments passed to the transform.

        Returns:
            A PyTree of transformed particles with strucutre $P$.

        Raises:
            AssertionError: if any of the transform properties are not valid/obeyed.
        """

        def validated_transform(
            particles: PyTree[Float[ArrayLike, "..."]]
        ) -> PyTree[Float[Array, "..."]]:
            """Allows for tree_map to be applied only over the particles."""
            _validated_transform = self.validate(self.transform)
            return _validated_transform(time, particles, args)

        return jtu.tree_map(validated_transform, particles, is_leaf=eqx.is_array_like)


class Component(AbstractComponent):
    """Defines a component without the need for subclassing `AbstractComponent`.

    Attributes:
        transformer: callable representing the component transform.
        validator: callable representing the component transform validator.
    """

    transformer: Callable[[float, PyTree, PyTree], PyTree] = identity_transform
    validator: Callable[[Callable[P, T]], Callable[P, T]] = identity_validator

    def transform(
        self, time: float, particles: PyTree[Float[ArrayLike, "..."]], args: PyTree
    ) -> PyTree[Float[Array, "..."]]:
        return self.transformer(time, particles, args)

    def validate(self, transform: Callable[P, T]) -> Callable[P, T]:
        return self.validator(transform)



r"""Standard components useful in the construction of MCCubature rules.

The primary purpose of the components API is to provide a convenient object-oriented 
interface for defining arbitrary mathematical transformations that are expected to obey 
certain transformation properties. In this package, these benefits are realized through:
    
* **Type Hinting**: when a callable is expected to have a certain call signature, and to 
  obey certain transformation properties, a suitable component is used as a type hint.

* **Transform Validation**: components consist of a transform and a validate method. The 
  validate method wraps the transform and allows for trace time validation of most
  transform properties (via Chex).

The latter case is particularly useful for developers wishing to implement new
specialized components as subclasses of existing ones, as they need only be concerned 
with defining a transform method that satisfies the existing validate method.

.. admonition:: What if I want a purely functional approach?
  :class: tip

  There is no obligation to use the component classes for any of the core functionality 
  in this package. If you want a purely functional approach, you can simply provide a 
  suitable callable (with a call and return signature identical to the expected 
  Component). However, in this case, the validity of the callable (with respect to the 
  expected component) is not checked. 

In the most general sense, a component can be mathematically defined as the composition 
$h = v \circ f$ of a transformation function $f: (t, p(t), args) \to p^{\prime}(t)$, 
where $t \in \mathbb{R}$ is a particle existence time, $p(t) \in \mathbb{R}^{n \times d}$ 
is the state of $n$ $d\text{-dimensional}$ particles at time $t$, $args$ is a set 
of arbitrary values that parameterise $f$, and $p^{\prime}(t) \in \mathbb{R}^{m \times d}$
are the transformed state of $m$ $d\text{-dimensional}$ particles; and a validation 
operator $v: f \to \{f, \varnothing\}$, with $\varnothing$ only returned if $f$ is not 
a valid transformation.

Note: 
  The particles could be considered as either a rank-two tensor ($n \times d$ matrix) or 
  as a set of $n$ rank-one tensors ($d\text{-vectors}$ $p(t) = \{p_i : p \in \mathbb{R^d}, i=1,\dots,n\}$.
  In this package the matrix form is always assumed unless explicitly stated otherwise.

Warning:
  Some specialized components modify the call-signature/domain of the transformation $f$
  and subsequently the composition $h$, from that presented above. In addition, while 
  all components are documented as accepting PyTrees of inputs, at present PyTrees are
  only fully supported for the specialized components in :mod:`mccube.components.recombinators`.
"""  # noqa: E501

from mccube.components.base import AbstractComponent, Component
from mccube.components.propagators import (
    AbstractPropagator,
    WrappedPropagator,
    LangevinDiffusionPropagator,
)
from mccube.components.recombinators import (
    AbstractRecombinator,
    WrappedRecombinator,
    MonteCarloRecombinator,
)


__all__ = [
    "AbstractComponent",
    "Component",
    "AbstractPropagator",
    "WrappedPropagator",
    "LangevinDiffusionPropagator",
    "AbstractRecombinator",
    "WrappedRecombinator",
    "MonteCarloRecombinator",
]



"""Recombinators that depend on random subsampling."""
from __future__ import annotations

from typing import Callable

import jax
import jax.tree_util as jtu
from jax._src.random import KeyArray
from jaxtyping import PyTree

from mccube.components.recombinators.base import AbstractRecombinator
from mccube.utils import no_operation, split_by_tree


class MonteCarloRecombinator(AbstractRecombinator):
    """Monte-carlo particle sub-sampling/recombination.

    Attributes:
        key: random key to initialise monte-carlo selection.
        weighting_function: selection weight/probability for each particle $x \in p(t)$.
        with_replacement: if to sample particles $x \in p(t)$ with/without replacement.
            Without replacement is usually significantly slower.
    """

    key: KeyArray
    weighting_function: Callable = no_operation
    with_replacement: bool = True

    def transform(
        self,
        key: KeyArray,
        recombination_factor: int | float,
        time: float,
        particles: PyTree,
        args: PyTree,
    ) -> PyTree:
        recombined_point_count = particles.shape[0] // recombination_factor
        weights = self.weighting_function(particles)
        return jax.random.choice(
            key,
            particles,
            (recombined_point_count,),
            self.with_replacement,
            weights,
        )

    def __call__(
        self,
        recombination_factor: Callable[[float, PyTree, PyTree], PyTree],
        time: float,
        particles: PyTree,
        args: PyTree,
    ) -> PyTree:
        key = jax.random.fold_in(self.key, time)
        keys = split_by_tree(key, particles)
        return jtu.tree_map(
            lambda key, particles: self.transform(
                key, recombination_factor, time, particles, args
            ),
            keys,
            particles,
        )



"""Base classes for defining and using Recombinator components."""
from __future__ import annotations

import functools
from typing import Callable, Sequence, overload

import chex
import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, ArrayLike, Float, PyTree

from mccube.components.base import AbstractComponent, P, T


class AbstractRecombinator(AbstractComponent):
    r"""Abstract base class for all Recombinators.

    Extends the concept of Components to transforms that depend on a recombination
    factor $r_f$. Also implements a concrete validate method to enforce the following
    properties of the transformation:

    1. The transformation **must return** the recombinant particles $p^{\prime}(t)$ as a **rank-two tensor** (matrix).
    2. The transformation **must not change the particle dimension** $d$.
    3. The transformation **must not increase the particle count** ($m \ge n$).
    """  # noqa: E501

    @overload
    def transform(
        self,
        recombination_factor: int | float,
        time: float,
        particles: PyTree[Float[ArrayLike, "n d"]],
        args: PyTree,
    ) -> PyTree[Float[Array, "m d"]]:
        r"""Transform the particles.

        Represents a function $f(r_f, t, p(t), args)$.

        Args:
            recombination_factor: factor by which to reduce the number of particles.
            time: particle existence time.
            particles: particles to transform.
            args: additional static arguments passed to the transform.

        Returns:
            Transformed/recombinant particles.
        """
        ...

    @staticmethod
    def validate(transform: Callable[P, T]) -> Callable[P, T]:
        r"""Validate the transform.

        Ensures that the transform obeys the properties:

        1. The transformation **must return** the recombinant particles $p^{\prime}(t)$ as a **rank-two tensor** (matrix).
        2. The transformation **must not change the particle dimension** $d$.
        3. The transformation **must not increase the particle count** ($m \ge n$).

        Args:
            transform: transform to validate, $f(r_f, t, p(t), args)$.

        Returns:
            Valid transform $(v \circ f)(r_f, t, p(t), args)$.
        """  # noqa: E501

        @functools.wraps(transform)
        def valid_transform(
            recombination_factor: int,
            time: float,
            particles: PyTree[Float[ArrayLike, "n d"]],
            args: PyTree,
        ) -> PyTree[Float[ArrayLike, "m d"]]:
            recombinant_particles = transform(
                recombination_factor, time, particles, args
            )
            chex.assert_rank(
                recombinant_particles,
                2,
                custom_message=(
                    "Recombinator transform must produce a particle array of rank two."
                ),
            )
            chex.assert_tree_shape_suffix(
                recombinant_particles,
                (particles.shape[-1],),
                custom_message=(
                    "Recombinator transform must not alter particle dimensionality."
                ),
            ),
            chex.assert_axis_dimension_lteq(
                jnp.vstack(recombinant_particles),
                0,
                particles.shape[0],
                custom_message=(
                    "Recombinator transform must not increase the number of particles."
                ),
            )

            return recombinant_particles

        return valid_transform

    def __call__(
        self,
        recombination_factor: Callable[[float, PyTree, PyTree], PyTree],
        time: float,
        particles: PyTree[Float[ArrayLike, "n d"]],
        args: PyTree,
    ) -> PyTree[Float[ArrayLike, "m d"]]:
        r"""Evaluate valid transform, $(v \circ f)(r_f, t, p(t), args)$.

        Args:
            recombination_factor: factor by which to reduce the number of particles.
            time: particle existence time.
            particles: particles to transform with PyTree structure $P$.
            args: additional static arguments passed to the transform.

        Returns:
            A PyTree of transformed particles with structure $P$.

        Raises:
            AssertionError: if any of the transform properties are not valid/obeyed.
        """

        def validated_transform(
            particles: PyTree[Float[ArrayLike, "n d"]]
        ) -> PyTree[Float[Array, "m d"]]:
            """Allows for tree_map to be applied only over the particles."""
            _validated_transform = self.validate(self.transform)
            return _validated_transform(recombination_factor, time, particles, args)

        return jtu.tree_map(validated_transform, particles, is_leaf=eqx.is_array_like)


class WrappedRecombinator(AbstractRecombinator):
    """Wraps a sequence of Callables as a Recombinator.

    Callables are composed in the sequence [a, b, c] -> c(b(a(...))), and are all
    assumed to have the same call signature as an `AbstractRecombinator`.
    Useful in cases where one has a suitable Callable, or composable sequence of
    Callables, which they wish to be validated and distinguished as an instance of
    :class:`AbstractRecombinator`.

    Attributes:
        recombinators: callables representing the Recombinator transforms.
    """

    # fmt: off
    recombinators: Sequence[Callable[P,T]] = lambda c, t, p, a: p  # noqa: E731
    # fmt: on
    def transform(
        self,
        recombination_factor: int | float,
        time: float,
        particles: PyTree[Float[ArrayLike, "n d"]],
        args: PyTree,
    ) -> PyTree[Float[ArrayLike, "m d"]]:
        return _compose_recombinators(self.recombinators)(
            recombination_factor, time, particles, args
        )


def _compose_recombinators(*transforms):
    def compose(f, g):
        return lambda ld, t, p, args: g(ld, t, f(t, p, args), args)

    if isinstance(transforms, Sequence) and len(transforms) > 1:
        return functools.reduce(compose, transforms)
    return transforms[0]



from __future__ import annotations

import abc
from typing import Callable, Sequence, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
from jaxtyping import Array, ArrayLike, Float, Int, PyTree
from scipy.linalg import hadamard as scipy_hadamard

# TODO: include the error estimates for each cubature formula where available.
# TODO: include an integration function.
# TODO: consider moving into separate package.


class _CubatureFormulaRegistryMeta(eqx._module._ModuleMeta):
    _registry = set()

    def __init__(cls, name, bases, attrs):
        super(_CubatureFormulaRegistryMeta, cls).__init__(name, bases, attrs)
        if name not in ["AbstractCubatureFormula", "_InitableModule"]:
            _CubatureFormulaRegistryMeta._registry.add(cls)

    @classmethod
    def get_registered_formulae(cls):
        return cls._registry


formula_registry = _CubatureFormulaRegistryMeta


class AbstractIntegrationRegion(eqx.Module):
    dimension: int

    @abc.abstractmethod
    def weight_function(x: ArrayLike) -> Array:
        ...

    @abc.abstractproperty
    def volume(self) -> float:
        ...


class EuclideanIntegrationRegion(AbstractIntegrationRegion):
    """N-dimensional Euclidean integration region, $\mathbb{R}^n$."""

    def weight_function(self, x: ArrayLike) -> Array:
        return 1.0

    @property
    def volume(self) -> float:
        return np.inf


class GaussianIntegrationRegion(AbstractIntegrationRegion):
    """N-dimensional unnormalized gaussian weighted Euclidean integration region.

    Notated as $E_n^{r^2}$ in :cite:p:`stroud1971`, this region represents integration
    over an N-dimensional Euclidean space, weighted by $\exp{-x_1^2 - x_2^2 \dots -x_d^2}$.
    """  # noqa: E501

    def weight_function(self, x: ArrayLike) -> Array:
        return np.exp(-np.sum(x**2, axis=-1))

    @property
    def volume(self) -> float:
        return np.pi ** (self.dimension / 2)


class AbstractCubatureFormula(eqx.Module, metaclass=_CubatureFormulaRegistryMeta):
    r"""Abstract base class for cubature formulae.

    Attributes:
        dimension:

        degree: polynomial degree $D$ for which the cubature formula is an exact
            integrator.
        sparse: if cubature formula produces a sparse matrix of cubature vectors.
        valid_dimensions: a range or specific set of vector-filed dimensions $d$, for
            which the cubature is an exact integrator of polynomials of degree $D$.
    """
    dimension: int
    degree: int = 0
    sparse: bool = False
    region: AbstractIntegrationRegion = EuclideanIntegrationRegion
    valid_dimensions: Tuple[int, int] | Sequence[Int] = (0, np.inf)

    def __init__(self, dimension):
        self.dimension = self._validate_dimension(dimension) and dimension
        self.region = self.region(self.dimension)

    @classmethod
    def _validate_dimension(cls, dimension: int) -> bool:
        """If cubature formula is valid for the given dimension.

        Args:
            dimension: dimension of the vector-field over which the cubature formula
                should be an exact integrator.

        Raises:
            ValueError: dimension not in range/set specified by `self.valid_dimensions`.
        """
        valid_dims = cls.valid_dimensions
        min_dim, max_dim = min(valid_dims), max(valid_dims)
        in_range = min_dim <= dimension <= max_dim
        # If valid_dims is length 2, then treat as range (lower, upper), else treat as
        # list/sequence of allowed dimensions.
        in_list = dimension in cls.valid_dimensions if len(valid_dims) < 2 else True
        if in_range and in_list:
            return True
        in_range_err_msg = f"{min_dim} <= dimension <= {max_dim}"
        in_list_err_msg = f"dimension in {valid_dims}"
        err_msg = in_range_err_msg if len(valid_dims) < 2 else in_list_err_msg
        raise ValueError(
            "Cubature formula cannot exactly integrate a vector-field of",
            f"the give dimension; expected {err_msg}, got {dimension}",
        )

    @abc.abstractproperty
    def coefficients(self) -> PyTree[float]:
        ...

    @abc.abstractproperty
    def vector_count(self) -> int:
        ...

    @abc.abstractproperty
    def vectors(self) -> PyTree[Float[Array, "k d"]]:
        ...

    def __call__(self, func: Callable[[Float[ArrayLike, " d"]], Float[Array, ""]]):
        """Approximately integrate some function $f$ over the integration region.

        Args:
            f: the jax transformable function to integrate.

        Returns:
            Approximated integral and weighted evaluations of $f$ at each vector $e_i$.
        """
        func_vmap = jax.vmap(func, 0)
        eval_points = jnp.vstack(
            jtu.tree_map(lambda c, v: c * func_vmap(v), self.coefficients, self.vectors)
        )
        return sum(eval_points), eval_points


class AbstractGaussianCubatureFormula(AbstractCubatureFormula):
    """Cubature formulae for a :class:`GaussianIntegrationRegion`."""

    region: AbstractIntegrationRegion = GaussianIntegrationRegion


# TODO: add reference here.
class Hadamard(AbstractGaussianCubatureFormula):
    degree: int = 4

    @property
    def coefficients(self) -> float:
        return self.region.volume / self.vector_count

    @property
    def vector_count(self) -> int:
        max_power = np.ceil(np.log2(self.dimension))
        max_dim = np.power(2, max_power)
        return max_dim * 2

    @property
    def vectors(self) -> Int[Array, "k d"]:
        hadamard_matrix = scipy_hadamard(self.vector_count // 2)
        new_matrix = hadamard_matrix[:, : self.dimension]
        return np.vstack((new_matrix, -new_matrix))


class StroudSecrest63_31(AbstractGaussianCubatureFormula):
    r"""Degree 3 cubature formula from :cite:p:`stroudSecrest1963`, listing $E^{r^2}_n$
    3-1. (pg315) in :cite:p:`stroud1971`."""

    degree: int = 3
    sparse: bool = True

    @property
    def coefficients(self) -> float:
        return self.region.volume / self.vector_count

    @property
    def vector_count(self) -> int:
        return 2 * self.dimension

    @property
    def vectors(self) -> Float[Array, "k d"]:
        radius = np.sqrt(self.dimension / 2)
        points_symmetric = radius * np.diag(np.ones(self.dimension))
        points_fully_symmetric = np.vstack([points_symmetric, -points_symmetric])
        return points_fully_symmetric * np.sqrt(2)


def minimal_cubature_formula(
    dimension: int,
    degree: int = 3,
    region=GaussianIntegrationRegion,
    sparse_only: bool = False,
    return_all: bool = False,
) -> AbstractCubatureFormula | Sequence[AbstractCubatureFormula]:
    """Cubature formula with the fewest cubature vectors for a given degree.

    Args:
        dimension: dimension of the space over which the cubature matrix must integrate.
        degree: polynomial degree for which the cubature matrix is an exact integrator.
        sparse_only: only consider cubature formulas that produce sparse vectors.
        return_all: if multiple
    """
    formulae = formula_registry.get_registered_formulae()
    if sparse_only:
        formulae = {f for f in formulae if f.sparse}
    # Need to convert set to array for indexing only the lowest count formulae.
    valid_formulae = np.fromiter(
        (
            f(dimension)
            for f in formulae
            if f.degree >= degree and issubclass(f.region, region)
        ),
        dtype=np.object_,
    )
    formulae_points = np.fromiter((f.vector_count for f in valid_formulae), dtype=int)
    minimal_formulae = np.where(formulae_points == formulae_points.min())
    minimal_formulae = valid_formulae[minimal_formulae]

    if return_all:
        return minimal_formulae
    return minimal_formulae[0]



r"""Components that are discrete time-integrators.

Propagators are specialized components that accept a modified call signature of the form 
$h(g, t, p(t), args)$, where $g: (t, p(t), args) \to p^{\prime}(t)$ represents some 
standard component/suitable-callable, and which have a validator $v$ which ensures the 
transformation $f$ obeys the following properties:

1. The transformation **must return** the propagated particles $p^{\prime}(t)$ as a **rank-two tensor** (matrix).
2. The transformation **must not change the particle dimension** $d$. 
3. The transformation **must not reduce the particle count** ($m \ge n$).

.. admonition:: The importance of propagators in cubature
    :class: important

    A Propagator can be considered as a discrete time-integration step for a (stochastic)
    differential equation or some integral $\int h(g, t, p(t), args) \mathrm{d}W(t)$. 
    In this package, the integrand $h$ is usually an SDE of a fixed form, $g$ is an 
    interaction potential (log-density function), and $W(t)$ is a collection of one or 
    more Brownian motion paths, where each concrete implementation of a Propagator 
    implicitly specifies the form of $h$ and $W(t)$ via its `transform` method.
"""  # noqa: E501

from mccube.components.propagators.base import AbstractPropagator, WrappedPropagator
from mccube.components.propagators.diffusions import LangevinDiffusionPropagator

__all__ = ["AbstractPropagator", "WrappedPropagator", "LangevinDiffusionPropagator"]



r"""Components that are measure-integration respecting maps on discrete measures.

Recombinators are specialized components that accept a modified call signature 
$h(r_f, t, p(t), args)$, where $r_f \in \mathbb{R}$ represents some recombination factor, 
and which have a validator $v$ which ensures the transformation $f$ obeys the following 
properties:

1. The transformation **must return** the recombinant particles $p^{\prime}(t)$ as a **rank-two tensor** (matrix).
2. The transformation **must not change the particle dimension** $d$. 
3. The transformation **must not increase the particle count** ($m \ge n$).

.. admonition:: Is a Recombinator just a Compressor?
    :class: note

    A Recombinator is a strictly more general component than a Compressor, as it 
    doesn't imply that the transformation aims to minimize some information 
    discrepancy/loss.

.. admonition:: The importance of Recombinators in cubature
    :class: important

    A Recombinator can be considered as a map on discrete measures which respects
    measure-integration $\bigg|\int_{\Omega} f(x_i) d\mu(x_i) - \int_{\Omega} f(x_j) d\nu(x_j)\bigg|$
    where the elements of the support of the two measures can be disjoint, the
    support of $\nu$ has cardinality
    $\lfloor \text{card}(\text{supp}(\mu)) / r_f \rfloor \le \text{card}(\text{supp}(\mu))$,
    and $i$ and $j$ index each co-ordinate dimension of the particle $x$.
    The support of $\nu$ is the recombinant particles $p^{\prime}(t)$, and would ideally (but not 
    neccessarily) be selected to minimize the above integral metric, while the support 
    of $\mu$ is simply the input particles $p(t)$. Each concrete implementation of a 
    Recombinator implicitly specifies the form of $f$, $\nu$ and $\mu$ via its 
    :meth:`AbstractComponent.transform() <mccube.components.base.AbstractComponent.transform>` method.

A typical usage pattern in this package is to compose a Recombinator and a Propagator
(see :class:`MCCubatureStep <mccube.cubature.MCCubatureStep>`). This is important in cases 
where repeated application of a Propagator could lead to particle count explosion. In 
such a case, the Recombinator is usually designed to `best` represent the expanded 
Propagated particles, by minimizing some integral metric, such as the one presented above.
"""  # noqa: E501

from mccube.components.recombinators.base import (
    AbstractRecombinator,
    WrappedRecombinator,
)
from mccube.components.recombinators.random import MonteCarloRecombinator

__all__ = [
    "AbstractRecombinator",
    "WrappedRecombinator",
    "MonteCarloRecombinator",
]



"""Discrete time-integration step for specific diffusion processes (fixed form SDEs)."""

from __future__ import annotations

from typing import Callable

import jax
import jax.numpy as jnp
from jaxtyping import Array, Float, PyTree

from mccube.components.propagators.base import AbstractPropagator


class LangevinDiffusionPropagator(AbstractPropagator):
    r"""An overdamped Langevin diffusion SDE particle Propagator.

    Permits a stationary ergodic solution, that is proportional to $\exp{-f}$. Hence
    is useful for Bayesian inference/sampling from $f$.

    Equivalent to a single step of the Euler-Maruyama time-integration scheme for the
    overdamped Langevin equation $Y^{i}_{t+1} := Y_t -\nabla f(Y_t) h + \sqrt{2h} e_i$,
    where $\{e_i\}$ are elements of the matrix of cubature_vectors $M$, $h$ is the
    Euler-Maruyama `step_size`, and $f$ is a `logdensity`, parameterising the particle
    interaction potential, $Y_t := p(t)$, and $Y^{i}_{t+1} := p^{\prime}(t)$.

    Note that the dimension of the vectors $Y^i$ and the cubature vectors $e_i$ must be
    identical and that for any collection of `cubature_matrix` with $k > 1$,
    $\text{card}(\{Y^{i}_{t+1}\}) > \text{card}(\{Y_t\})$.

    Attributes:
        cubature_matrix: matrix of $d\text{-dimensional}$ cubature vectors.
        step_size: Euler-Maruyama step size $h$.
    """

    cubature_matrix: Float[Array, "k d"]
    step_size: float = 0.1

    def transform(
        self,
        logdensity: Callable[[float, PyTree, PyTree], PyTree],
        time: float,
        particles: PyTree,
        args: PyTree,
    ) -> PyTree:
        dim = particles.shape[-1]
        diffusion = jnp.sqrt(2.0 * self.step_size) * self.cubature_matrix
        drift = jax.vmap(jax.grad(logdensity, argnums=1), [None, 0, None])(
            time, particles, args
        )
        drift *= self.step_size
        return (particles + drift + diffusion[:, None, :]).reshape(-1, dim)



import abc
import os
import time
from queue import Queue
from threading import Thread

import jax.numpy as jnp
from jaxtyping import Float, Array
from mccube.cubature import MCCubatureState
from mccube.metrics import cubature_target_error
from mccube.utils import no_operation
from tensorboardX import SummaryWriter


class AbstractMCCubatureVisualizer:
    """Defines a Markov chain cubature visualization step."""

    @abc.abstractmethod
    def __call__(self, state: MCCubatureState):
        ...


class NonblockingVisualizer(AbstractMCCubatureVisualizer):
    """Defines a nonblocking MCCubature visualizer.

    Attributes:
        logging_interval: interval at which to perform visualization/logging.
        target_mean: mean against which to compare the MCCubature mean.
        target_cov: covariance against which to compare the MCCubature covariance.
    """

    def __init__(
        self,
        logging_interval: int = 1,
        target_mean: Float[Array, " d"] = None,
        target_cov: Float[Array, "d d"] = None,
    ):
        """Initialise a cubature visualizer.

        Args:
            logging_interval: interval at which to perform visualization/logging.
            target_mean: mean against which to compare the MCCubature mean.
            target_cov: covariance against which to compare the MCCubature covariance.
        """
        self.logging_interval = logging_interval
        self.target_mean = target_mean
        self.target_cov = target_cov
        self._queue = Queue()

    def __call__(self, state: MCCubatureState, *args, **kwargs):
        """Visulization callback to be passed to the MCCubatureStep."""
        epoch = state.time
        particles = state.particles
        current_time = time.time()
        if not hasattr(self, "start_time_"):
            self.start_time_ = current_time
        if epoch % self.logging_interval:
            return
        self.elapsed_time_ = current_time - self.start_time_
        self._queue.put((particles, epoch, current_time))

    def __enter__(self):
        self.start_visualization()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stop_visualization()

    def _inprocess_visualization(self, visualization_function=no_operation):
        """Visualization operations to perform on queued cubature steps."""
        while True:
            particles, epoch, walltime = self._queue.get()
            self.epoch_ = epoch
            visualization_function(particles, epoch, walltime)
            if self.target_mean and self.target_cov:
                self.error_ = cubature_target_error(
                    particles, self.target_mean, self.target_cov
                )
            self._queue.task_done()

    def _postprocess_visualization(self):
        """Visualization to perform after the final cubature epoch has terminated."""
        string = f"""
        Visualization Complete
        Experiment Time: {self.elapsed_time_}
        """
        if self.target_mean and self.target_cov:
            normalized_mean_error = self.error_[0].l2 / jnp.linalg.norm(
                self.target_mean
            )
            normalized_variance_error = self.error_[1].l2 / jnp.linalg.norm(
                jnp.diag(self.target_cov)
            )
            string += f"""
            Error Mean: {self.error_[0].l2} ({100 * 1 - normalized_mean_error} %)
            Error Var: {self.error_[1].l2} ({100 * 1 - normalized_variance_error} %)
            """
        print(string)

    def start_visualization(self):
        """Start the visualization thread."""
        self._visualizer_thread = Thread(
            target=self._inprocess_visualization, daemon=True
        )
        self._visualizer_thread.start()
        # May need a finalizer for the case where stop_visualizer isnt called?

    def stop_visualization(self):
        """Finish processing items in the queue and stop the visualization thread."""
        if self._queue.not_empty:
            print("Visualiser is finishing up...")
            self._queue.join()
        self._postprocess_visualization()


class TensorboardVisualizer(NonblockingVisualizer):
    """Tensorboard compatible cubature visualizer.

    Attributes:
        logdir: directory in which to store the tensorboard logs.
        logging_interval: interval at which to perform visualization/logging.
        hparams: list of cubature hyper-parameters.
        plot_figures: indicates if to plot intermediate cubature step particles.
    """

    def __init__(
        self,
        logdir=None,
        logging_interval=1,
        hparams=None,
        experiment_name=str(time.time()),
        **kwargs,
    ):
        """Initialises a Tensorboard cubature visualizer.

        Args:
            logdir: directory in which to store the tensorboard logs.
            logging_interval: interval at which to perform visualization/logging.
            hparams: list of cubature hyper-parameters.
            experiment_name: name of the experimental run.
            plot_figures: indicates if to plot intermediate cubature step particles.
        """
        self.logdir = os.path.join(logdir or os.getcwd(), experiment_name)
        self.logging_interval = logging_interval
        self.hparams = hparams
        super().__init__(logging_interval=logging_interval, **kwargs)

    def _inprocess_visualization(self):
        def visualisation_function(particles, epoch, walltime):
            if self.target_mean and self.target_cov:
                self.error_ = cubature_target_error(
                    particles, self.target_mean, self.target_cov
                )
                self._writer.add_scalar(
                    tag="Error/Mean",
                    scalar_value=self.error_[0].l2,
                    global_step=epoch,
                    walltime=walltime,
                )
                self._writer.add_scalar(
                    tag="Error/Variance",
                    scalar_value=self.error_[1].l2,
                    global_step=epoch,
                    walltime=walltime,
                )
            self._writer.add_scalar(
                tag="Speed/Time_per_epoch",
                scalar_value=self.elapsed_time_ / (epoch or 1),
                global_step=epoch,
                walltime=walltime,
            )
            self._writer.add_histogram(
                tag="Distribution/ProjectedHistogram",
                values=particles,
                global_step=epoch,
                walltime=walltime,
            )

        self._writer = SummaryWriter(self.logdir)
        super()._inprocess_visualization(visualization_function=visualisation_function)

    def _postprocess_visualization(self):
        if self.hparams:
            metric_dict = {
                "Speed/Elapsed_time": self.elapsed_time_,
                "Speed/Time_per_epoch": self.elapsed_time_ / (self.epoch_ or 1),
            }
            if self.target_mean and self.target_cov:
                metric_dict.update(
                    {
                        "Error/Mean": self.error_[0].l2,
                        "Error/Variance": self.error_[1].l2,
                    }
                )
            self._writer.add_hparams(
                hparam_dict=self.hparams,
                metric_dict=metric_dict,
                name=".",
                global_step=self.epoch_,
            )
        self._writer.close()
        super()._postprocess_visualization()



"""Markov Chain Cubature (MCC) construction/inference/solving.

This module provides the tools for constructing SDE cubatures via MCC.

An SDE cubature (alternatively called a Cubature on Wiener Space :cite:p:`lyons2004`) 
is an approximate discrete-time integrator that exactly integrates weak-solutions 
(moments up to degree $N$) for some class of SDEs. This works by reducing the problem 
of integrating the SDE with respect to its infinitely many strong solution paths, to 
obtain some expected path, to one of taking discrete sums of finitely many, so called 
cubature paths.

In measure terminology, an SDE cubature replaces the SDE's Wiener measure with a 
discrete measure, supported on finitely many bounded variation paths, called a cubature 
measure/cubature paths (essentially reducing the problem to one of solving many ODEs 
instead of a single SDE). See :cite:t:`crisan2017cubature` for a nice introduction to 
the method, and :cite:t:`lyons2004` for a comprehensive investigation.

The salient limitation of SDE cubatures constructed as per :cite:t:`lyons2004`
is that the paths count scales exponentially with the number of discrete time steps 
($\mathcal{O}(n^{m})$, where $n$ is the propagator expansion factor, and $m$ is the 
number of time-integration steps). MCC solves this problem by constructing the 
collection of paths as a markov chain, wherethe :class:`MCCubatureStep` acts as a 
transition kernel that employs recombination to maintain the path/particle count at 
every time step. Note that in MCCube the paths are usually interpreted as particle 
trajectories, as this provides a consistent physically analogy.
"""

from __future__ import annotations

from typing import Callable

import chex
import equinox as eqx
import jax
from jax.experimental.host_callback import id_tap
from jaxtyping import Array, ArrayLike, Float, PyTree

from mccube.components import (
    AbstractPropagator,
    AbstractRecombinator,
    WrappedPropagator,
    WrappedRecombinator,
)
from mccube.utils import no_operation


class MCCubatureState(eqx.Module):
    time: float
    particles: PyTree[Float[Array, "n d"]]
    args: PyTree


class MCCubatureStep(AbstractPropagator):
    r"""Shape preserving (non-expanding) MCCubature step/transition kernel.

    `MCCubatureStep` automatically computes the recombination factor such that the
    composition of Propagator and Recombinator results in no change to particle count
    or dimension. Also implements a concrete validate method to enforce the following
    properties of the composed transformation:

    1. Particle PyTree structure is conserved.
    2. Particle count, rank, and dimensionality are conserved.

    Attributes:
        propagator: a Propagator component.
        recombinator: a Recombinator component.
    """
    propagator: AbstractPropagator = WrappedPropagator()
    recombinator: AbstractRecombinator = WrappedRecombinator()

    def transform(
        self,
        logdensity: Callable[[float, PyTree, PyTree], PyTree],
        time: float,
        particles: PyTree[Float[ArrayLike, "n d"]],
        args: PyTree,
    ) -> PyTree[Float[Array, "n d"]]:
        propagated_particles = self.propagator(logdensity, time, particles, args)
        recombination_factor = propagated_particles.shape[0] // particles.shape[0]
        recombinant_particles = self.recombinator(
            recombination_factor, time, propagated_particles, args
        )
        return recombinant_particles

    def validate(self, transform):
        r"""Validate the transform.

        Ensures that the transform obeys the properties:

        1. The transformation **must conserve the particle PyTree structure** $P$.
        2. The transformation **must conserve particle count and dimensionality,
        $p(t) \in \mathbb{R}^{n \times d}$ and $p^{\prime}(t) \in \mathbb{R}^{n \times d}$**.

        Args:
            transform: transform to validate, $h(g, t, p(t), args)$.

        Returns:
            Valid transform $(v \circ f)(g, t, p(t), args)$.
        """  # noqa: E501

        def valid_step(
            logdensity: Callable[[float, PyTree, PyTree], PyTree],
            time: float,
            particles: PyTree,
            args: PyTree,
        ) -> PyTree:
            updated_particles = transform(logdensity, time, particles, args)
            chex.assert_trees_all_equal_shapes(
                updated_particles,
                particles,
                custom_message=(
                    "Cubature step must return a particle PyTree with the same shape as"
                    " the input particle PyTree."
                ),
            )
            return updated_particles

        return valid_step


def mccubaturesolve(
    logdensity: Callable[[float, PyTree, PyTree], PyTree],
    transition_kernel: MCCubatureStep | AbstractPropagator,
    initial_particles: Float[Array, "n d"],
    args: PyTree | None = None,
    epochs: int = 500,
    *,
    t0: float = 0.0,
    dt: float = 1.0,
    visualization_callback: Callable[[MCCubatureState], None] | None = None,
) -> PyTree[MCCubatureState]:
    r"""Fixed-timestep MCCubature construction/inference/solution loop.

    Constructs and evaluates an MCCubature by iterative application of a
    `transition_kernel`/:class:`MCCubatureStep`.

    For a suitable SDE (Propagator), :func:`mccubaturesolve` generates a collection of
    paths as a discrete-time Markov chain, where the `MCCubatureStep` defines the state
    transition kernel $h(g, t, p(t), args)$, `logdensity` defines the (unnormalized)
    density of the distribution $g(t, p(t), args)$ that parametrizes the SDE, `epochs`
    defines the number of discrete time steps $dt$, and `prior_samples` defines the
    chain starting state $p(t0)$.

    When used for Bayesian inference (for example when the SDE/Propagator is the
    :class:`LangevinDiffusionPropagator`), the initial state of the chain $p(t_0)$
    is the set of prior samples, the logdensity $g$ is consistent with the posterior
    (the distribution one wishes to sample from), and the final state of the chain
    $p(t_0 + dt * \text{epochs})$ is the set of posterior samples.W

    See :mod:`mccube.extensions.solvers` for a more general approach to solving
    SDEs via MCCubature.

    Args:
        logdensity: the log-density of the distribution that parametrizes the
            SDE/Propagator (in the case of Bayesian inference this is the distribution
            one wishes to sample from).
        transition_kernel: the MCCubature/integration step.
        epochs: the number of cubature steps (discrete timesteps) to take.
        initial_particles: the particles to use as the initial state of the Markov chain
            (in the case of Bayesian inference these are samples from the prior).
        args: additional static arguments passed to the log-density function.
        t0: the timestamp of the seed event for the Markov Chain.
        dt: the fixed-timestep between events in the Markov Chain.
        visualization_callback: can be used to perform in-process visualisation of the
            `CubatureState` at each epoch.

    Returns:
        The :class:`MCCubatureState` at every timestep (collection of evaluated
        cubature paths).

        Those familiar with cubature formulae may wish to think of the particles in
        each :class:`MCCubatureState` as a collection of cubature vectors $v_i$, whose
        cubature coefficients are all $B_i = 1/n$. For weak quantities such as the mean,
        \int x dP(x) \approx \frac{1}{n} \sum_i^{n}, the discrete estimator is identical
        to the implied cubature formula $\sum_i^{n} B_i f(v)$, for $f(v) = v$.
    """
    initial_particles = MCCubatureState(time=t0, particles=initial_particles, args=args)
    visualization_callback = visualization_callback or no_operation  # Handle None case
    id_tap(visualization_callback, initial_particles)

    def body_fun(state, _):
        time = state.time
        particles = state.particles
        updated_particles = transition_kernel(logdensity, time, particles, args)
        updated_state = eqx.tree_at(
            lambda state: [state.time, state.particles],
            state,
            [time + dt, updated_particles],
        )
        id_tap(visualization_callback, updated_state)
        return updated_state, updated_state

    _, final_state = jax.lax.scan(body_fun, initial_particles, None, length=epochs)
    return final_state



from typing import Tuple

import equinox as eqx
import numpy as np
from jaxtyping import ArrayLike, Float


class ResultStatistics(eqx.Module):
    value: ArrayLike
    maximum: float
    minimum: float
    dimension_averaged: ArrayLike
    l2: float


def _error_stats(a, b):
    abserr = np.abs(a - b)
    l2err = np.linalg.norm(a - b)
    return ResultStatistics(a, np.max(abserr), np.min(abserr), np.mean(abserr), l2err)


def cubature_target_error(
    particles: Float[ArrayLike, "n d"],
    target_mean: Float[ArrayLike, " d"],
    target_covariance: Float[ArrayLike, "d d"],
) -> Tuple[ResultStatistics, ResultStatistics]:
    mean_err = _error_stats(np.mean(particles, axis=0), target_mean)
    cov_err = _error_stats(np.cov(particles, rowvar=False), target_covariance)
    return (mean_err, cov_err)



"""Base classes for defining and using Propagator components."""
from __future__ import annotations

import functools
from typing import Callable, Sequence, overload

import chex
import equinox as eqx
import jax.tree_util as jtu
from jaxtyping import Array, ArrayLike, Float, PyTree

from mccube.components.base import AbstractComponent, P, T


class AbstractPropagator(AbstractComponent):
    r"""Abstract base class for all Propagators.

    Extends the concept of Components to transforms that depend on the evaluation
    of a function $g(t, p(t), args)$. Also implements a concrete validate method to
    enforce the following properties of the transformation:

    1. The transformation **must return** the propagated particles $p^{\prime}(t)$ as a **rank-two tensor** (matrix).
    2. The transformation **must not change the particle dimension** $d$.
    3. The transformation **must not reduce the particle count** ($m \ge n$).
    """  # noqa: E501

    @overload
    def transform(
        self,
        logdensity: Callable[[float, PyTree, PyTree], PyTree],
        time: float,
        particles: PyTree[Float[ArrayLike, "n d"]],
        args: PyTree,
    ) -> PyTree[Float[Array, "m d"]]:
        r"""Transform the particles.

        Represents a function $f(g, t, p(t), args)$.

        Args:
            logdensity: particle interaction log-density function $g(t, p(t), args)$.
            time: particle existance time.
            particles: particles to transform.
            args: additional static arguments passed to the transform.

        Returns:
            Transformed/propagated particles.
        """
        ...

    @staticmethod
    def validate(transform: Callable[P, T]) -> Callable[P, T]:
        r"""Validate the transform.

        Ensures that the transform obeys the properties:

        1. The transformation **must return** the propagated particles $p^{\prime}(t)$
        as a **rank-two tensor** (matrix).
        2. The transformation **must not change the particle dimension** $d$.
        3. The transformation **must not reduce the particle count** ($m \ge n$).

        Args:
            transform: transform to validate, $f(g, t, p(t), args)$.

        Returns:
            Valid transform $(v \circ f)(g, t, p(t), args)$.
        """

        @functools.wraps(transform)
        def valid_transform(
            logdensity: Callable[[float, PyTree, PyTree], PyTree],
            time: float,
            particles: PyTree[Float[ArrayLike, "n d"]],
            args: PyTree[Float[Array, "m d"]],
        ) -> PyTree:
            propagated_particles = transform(logdensity, time, particles, args)
            chex.assert_rank(
                propagated_particles,
                2,
                custom_message=(
                    "Propagator transform should return particle array of rank two."
                ),
            )
            chex.assert_equal_shape_suffix(
                [propagated_particles, particles],
                1,
                custom_message=(
                    "Propagator transform must not alter particle dimensionality."
                ),
            )
            chex.assert_axis_dimension_gteq(
                propagated_particles,
                0,
                particles.shape[0],
                custom_message=(
                    "Propagator transform must not reduce the number of particles."
                ),
            )
            return propagated_particles

        return valid_transform

    def __call__(
        self,
        logdensity: Callable[[float, PyTree, PyTree], PyTree],
        time: float,
        particles: PyTree[Float[ArrayLike, "n d"]],
        args: PyTree,
    ) -> PyTree[Float[Array, "m d"]]:
        r"""Evaluate valid transform, $(v \circ f)(g, t, p(t), args)$.

        Args:
            logdensity: particle interaction log-density function $g(t, p(t), args)$.
            time: particle existance time.
            particles: particles to transform with PyTree structure $P$.
            args: additional static arguments passed to the transform.

        Returns:
            A PyTree of transformed particles with structure $P$.

        Raises:
            AssertionError: if any of the transform properties are not valid/obeyed.
        """

        def validated_transform(
            particles: PyTree[Float[ArrayLike, "n d"]]
        ) -> PyTree[Float[Array, "m d"]]:
            """Allows for tree_map to be applied only over the particles."""
            _validated_transform = self.validate(self.transform)
            return _validated_transform(logdensity, time, particles, args)

        return jtu.tree_map(validated_transform, particles, is_leaf=eqx.is_array_like)


class WrappedPropagator(AbstractPropagator):
    """Wraps a sequence of Callables as a Propagator.

    Callables are composed in the sequence [a, b, c] -> c(b(a(...))), and are all
    assumed to have the same call signature as an :class:`AbstractPropagator`. Useful in
    cases where one has a suitable Callable, or composable sequence of Callables, which
    they wish to be validated and distinguished as an instance of
    :class:`AbstractPropagator`.

    Attributes:
        propagators: callables representing the Propagator transforms.
    """

    # fmt: off
    propagators: Sequence[Callable[[float, PyTree, PyTree], PyTree]] = lambda ld, t, p, a: p  # noqa: E731, E501
    # fmt: on
    def transform(
        self,
        logdensity: Callable[[float, PyTree, PyTree], PyTree],
        time: float,
        particles: PyTree[Float[ArrayLike, "n d"]],
        args: PyTree,
    ) -> PyTree[Float[ArrayLike, "m d"]]:
        return _compose_propagators(self.propagators)(logdensity, time, particles, args)


def _compose_propagators(*transforms):
    def compose(f, g):
        return lambda ld, t, p, args: g(ld, t, f(t, p, args), args)

    if isinstance(transforms, Sequence) and len(transforms) > 1:
        return functools.reduce(compose, transforms)
    return transforms[0]
